#include "pg.h"

static unsigned seed1;    
static unsigned seed2; 

/* Button bit masks */ 
#define CTRL_SQUARE     0x8000 
#define CTRL_TRIANGLE   0x1000 
#define CTRL_CIRCLE     0x2000 
#define CTRL_CROSS      0x4000 
#define CTRL_UP         0x0010 
#define CTRL_DOWN       0x0040 
#define CTRL_LEFT       0x0080 
#define CTRL_RIGHT      0x0020 
#define CTRL_START      0x0008 
#define CTRL_SELECT     0x0001 
#define CTRL_LTRIGGER   0x0100 
#define CTRL_RTRIGGER   0x0200 

/* Returned control data */ 
typedef struct _ctrl_data 
{ 
	int frame; 
	int buttons; 
	unsigned char analog[4]; 
	int unused; 
} ctrl_data_t; 

/* game structs and variables */
typedef struct
{
	int x;
	int y;
} Cell;

#define RGB(r, g, b) ((r)|((g)<<5)|((b)<<10))

#define BACKGROUND_COLOR RGB(0x1f, 0x1f, 0x1f)
#define CELL_COLOR RGB(0, 0x1f, 0)
#define TEXT_COLOR RGB(0, 0, 0)
#define FRAME_COLOR RGB(0, 0, 0x1f)
#define TARGET_COLOR RGB(0x1f, 0, 0)

Cell* cells = (Cell*)0x9300000;
Cell target;
int dx;
int dy;
int x;
int y;
int cellStart;
int cellEnd;
int maxCells = 60*34;
int score;
int high = 0;

void initSeeds() 
{ 
   seed1 = sceKernelLibcTime((void *) 0); 
   seed2 = sceKernelLibcTime((void *) 0); 
} 

unsigned int myRand() 
{ 
   seed1 = (seed1 + 46906481) ^ seed2; 
   seed2 = seed1 ^ ( ((seed2<<3) | (seed2 >> 29)) + 103995407); 
    
   return seed1 - seed2; 
}

int exit_callback(void)
{
	sceKernelExitGame();

	return 0;
}

int CallbackThread(void *arg)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback);
	sceKernelRegisterExitCallback(cbid);
	sceKernelSleepThreadCB();

	return 0;
}

int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}

void printDecimal(int x, int y, int color, int value)
{
	int i;
	int zero = 1;
	char digits[6];
	for (i = 5; i >= 0; i--) {
		digits[i] = '0' + (value%10);
		value /= 10;
	}
	i = 0;
	while (i < 5 && digits[i] == '0') digits[i++] = 0;
	for (i = 0; i < 6; i++) {
		if (digits[i]) {
			pgPutChar((x++)*8, y*8, color, BACKGROUND_COLOR, digits[i], 1, 1, 1);
		}
	}
}

void fillCell(int x, int y, int color)
{
	unsigned short* vram = (unsigned short*) pgGetVramAddr(x * 16, y * 16);
	int xo, yo;
	for (xo = 0; xo < 16; xo++) {
		for (yo = 0; yo < 16; yo++) {
			vram[xo + yo * 512] = color;
		}
	}
}

void createRandomTarget() {
	target.x = (myRand() % 21) + 1;
	target.y = (myRand() % 15) + 1;
}

void newGame()
{
	pgFlipDrawFrame();
	pgFillvram(BACKGROUND_COLOR);
	int i;
	for (i = 0; i < 30; i++) {
		fillCell(i, 0, FRAME_COLOR);
		fillCell(i, 16, FRAME_COLOR);
	}
	for (i = 1; i < 16; i++) {
		fillCell(0, i, FRAME_COLOR);
		fillCell(22, i, FRAME_COLOR);
		fillCell(29, i, FRAME_COLOR);
	}
	
	dx = 0;
	dy = 0;
	x = 12;
	y = 8;
	cells[0].x = x;
	cells[0].y = y;
	createRandomTarget();

	cellStart = 0;
	cellEnd = 0;
	pgPrint(49, 3, TEXT_COLOR, "Snake");
	pgPrint(47, 10, TEXT_COLOR, "Score:");
	pgPrint(47, 16, TEXT_COLOR, "Highscore:");
	pgPrint(48, 29, TEXT_COLOR, "by Shine");
	score = 0;
	pgFlipShowFrameV();
}

void keyboardControl()
{
	ctrl_data_t ctrl;
	sceCtrlReadBufferPositive(&ctrl, 1); 
	if (ctrl.buttons & CTRL_UP) { 
		dx = 0;
		dy = -1;
	} else if (ctrl.buttons & CTRL_DOWN) { 
		dx = 0;
		dy = 1;
	} else if (ctrl.buttons & CTRL_LEFT) { 
		dx = -1;
		dy = 0;
	} else if (ctrl.buttons & CTRL_RIGHT) { 
		dx = 1;
		dy = 0;
	}
}

void move()
{
	cellEnd++;
	if (cellEnd == maxCells) cellEnd = 0;
	cells[cellEnd].x = x;
	cells[cellEnd].y = y;
	if (x == target.x && y == target.y) {
		createRandomTarget();
		score++;
	} else {
		fillCell(cells[cellStart].x, cells[cellStart].y, BACKGROUND_COLOR);
		cellStart++;
		if (cellStart == maxCells) cellStart = 0;
	}
	fillCell(cells[cellEnd].x, cells[cellEnd].y, CELL_COLOR);
	x += dx;
	y += dy;
}

int isGameOver()
{
	int lastX = cells[cellEnd].x;
	int lastY = cells[cellEnd].y;
	int gameOver = 0;
	if (lastX >= 22) gameOver = 1;
	if (lastX < 1) gameOver = 1;
	if (lastY >= 16) gameOver = 1;
	if (lastY < 1) gameOver = 1;
	int start = cellStart;
	while (start != cellEnd) {
		if (cells[start].x == lastX && cells[start].y == lastY) {
			gameOver = 1;
			break;
		}
		start++;
		if (start == maxCells) start = 0;
	}
	return gameOver;
}
	
int xmain(int ra)
{
	initSeeds();
	SetupCallbacks();

	pgInit();
	pgScreenFrame(1, 1);
	
	newGame();
	
	while (1) {
		keyboardControl();
		pgWaitVn(5);
		move();
		fillCell(target.x, target.y, TARGET_COLOR);
		if (score > high) high = score;
		if (isGameOver()) {
			pgWaitVn(50);
			newGame();
		}
		printDecimal(47, 12, TEXT_COLOR, score);
		printDecimal(47, 18, TEXT_COLOR, high);
	}
	return 0;
}
